@@ -1,4 +1,4 @@
-/* Copyright (c) 2010, 2016, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2010, 2019, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -88,6 +88,21 @@ LEX_CSTRING event_names[][6] = {
 
 static volatile int number_of_calls;
 
+/*
+  Plugin has been installed.
+*/
+static my_bool g_plugin_installed= FALSE;
+
+/*
+  Record buffer mutex.
+*/
+static mysql_mutex_t g_record_buffer_mutex;
+
+/*
+  Event recording buffer.
+*/
+static char *g_record_buffer;
+
 #define AUDIT_NULL_VAR(x) static volatile int number_of_calls_ ## x;
 #include "audit_null_variables.h"
 #undef AUDIT_NULL_VAR
@@ -148,6 +163,7 @@ static MYSQL_THDVAR_STR(event_record_def,
                         "Event recording definition", NULL, NULL, NULL);
 
 static MYSQL_THDVAR_STR(event_record,
+                        PLUGIN_VAR_READONLY |
                         PLUGIN_VAR_RQCMDARG | PLUGIN_VAR_MEMALLOC,
                         "Event recording", NULL, NULL, NULL);
 /*
@@ -173,6 +189,13 @@ static int audit_null_plugin_init(void *arg MY_ATTRIBUTE((unused)))
     *((int*)var->value) = 0;
   }
 
+  mysql_mutex_init(PSI_NOT_INSTRUMENTED,
+                   &g_record_buffer_mutex,
+                   MY_MUTEX_INIT_FAST);
+
+  g_record_buffer= NULL;
+  g_plugin_installed= TRUE;
+
   return(0);
 }
 
@@ -192,6 +215,17 @@ static int audit_null_plugin_init(void *arg MY_ATTRIBUTE((unused)))
 
 static int audit_null_plugin_deinit(void *arg MY_ATTRIBUTE((unused)))
 {
+  if (g_plugin_installed == TRUE)
+  {
+    my_free((void *)(g_record_buffer));
+
+    g_record_buffer= NULL;
+
+    mysql_mutex_destroy(&g_record_buffer_mutex);
+
+    g_plugin_installed= FALSE;
+  }
+
   return(0);
 }
 
@@ -243,8 +277,8 @@ static LEX_CSTRING get_token(const char **str)
   return ret;
 }
 
-static void add_event(MYSQL_THD thd, const char *var, LEX_CSTRING event,
-                      const char *data, size_t data_length)
+static char *add_event(const char *var, LEX_CSTRING event,
+                       const char *data, size_t data_length)
 {
   LEX_CSTRING str;
   size_t size;
@@ -260,7 +294,7 @@ static void add_event(MYSQL_THD thd, const char *var, LEX_CSTRING event,
 
   buffer[size - (str.length == 0 ? 2 : 1)] = '\0';
 
-  THDVAR(thd, event_record)= buffer;
+  return buffer;
 }
 
 static void process_event_record(MYSQL_THD thd, LEX_CSTRING event_name,
@@ -281,10 +315,22 @@ static void process_event_record(MYSQL_THD thd, LEX_CSTRING event_name,
        a record variable */
 
     const char *buffer= THDVAR(thd, event_record);
+    char *new_buffer= NULL;
 
     /* Add event. */
-    add_event(thd, buffer, event_name, data, data_length);
+    mysql_mutex_lock(&g_record_buffer_mutex);
 
+    /* Only one THD is capable of adding events into the buffer. */
+    if (buffer && (buffer == g_record_buffer))
+    {
+      new_buffer= add_event(buffer, event_name, data, data_length);
+      g_record_buffer= new_buffer;
+      my_free((void *)(buffer));
+    }
+
+    mysql_mutex_unlock(&g_record_buffer_mutex);
+
+    THDVAR(thd, event_record)= new_buffer;
 
     if (!my_charset_latin1.coll->strnncoll(&my_charset_latin1,
                                            (const uchar *)record_begin.str,
@@ -298,6 +344,8 @@ static void process_event_record(MYSQL_THD thd, LEX_CSTRING event_name,
   }
   else
   {
+    const char *buffer;
+
     /* We have not started recording of events yet. */
     if (my_charset_latin1.coll->strnncoll(&my_charset_latin1,
                                           (const uchar *)record_begin.str,
@@ -309,9 +357,20 @@ static void process_event_record(MYSQL_THD thd, LEX_CSTRING event_name,
       return;
     }
 
-    THDVAR(thd, event_record)= 0;
+    buffer= THDVAR(thd, event_record);
 
-    add_event(thd, "", event_name, data, data_length);
+    mysql_mutex_lock(&g_record_buffer_mutex);
+
+    if (buffer == g_record_buffer)
+    {
+      my_free((void *)(buffer));
+
+      g_record_buffer= add_event("", event_name, data, data_length);
+
+      THDVAR(thd, event_record)= g_record_buffer;
+    }
+
+    mysql_mutex_unlock(&g_record_buffer_mutex);
 
     /* Add event. */
 
